import os
import re
from typing import Set
from shutil import rmtree, move
import sqlite3
import json
from multiprocessing import Pool
from pathlib import Path
import fire  # type: ignore
from tqdm import tqdm  # type: ignore
import rpyc  # type: ignore
from pywebcopy import WebPage, config  # type: ignore
from .persistor import Persistor
from .graph import graph
from .utils import remove_dir, remove_symlink, download_url


class HyperTag:
    """ HyperTag CLI """

    def __init__(self):
        self.db = Persistor()
        self.root_dir = Path(self.db.get_hypertagfs_dir())
        os.makedirs(self.root_dir, exist_ok=True)

    def search_image(
        self, *text_queries: str, cpu=None, top_k=10, path=0, score=0, verbose=0, _return=0
    ):
        """ Execute a semantic search that returns best matching images """
        text_query = " ".join(text_queries)
        try:
            rpc = rpyc.connect("localhost", 18861)
            results = rpc.root.search_image(text_query, path, top_k, score)
            for result in results:
                print(result)
            if len(result) == 0:
                print("No relevant files indexed...")
        except ConnectionRefusedError:
            from .vectorizer import CLIPVectorizer

            vectorizer = CLIPVectorizer(cpu, verbose)
            results = vectorizer.search_image(text_query, path, top_k, score)
        if _return:
            return results

    def index(self, text=None, image=None, rebuild=False, cache=False, cores: int = 0):
        """ Vectorize image & text files (needed for semantic search) """
        if (image and text is None) or (image and text) or (not image and not text):
            self.index_images()
        if (text and image is None) or (image and text) or (not image and not text):
            self.index_texts(rebuild, cache, cores)

    def index_images(self):
        """ Vectorize image files (needed for semantic search) """
        from .vectorizer import CLIPVectorizer, get_image_files

        file_paths = self.db.get_unvectorized_file_paths()
        compatible_files = get_image_files(file_paths, verbose=True)
        print("Vectorizing", len(compatible_files), "images...")
        remote = True
        if remote:
            try:
                rpc = rpyc.connect("localhost", 18861)
                rpc._config["sync_request_timeout"] = None  # Disable timeout
                print("Connected to DaemonService successfully")
            except ConnectionRefusedError:
                print("DaemonService connection failed, falling back to local execution...")
                remote = False

        if not remote:
            img_vectorizer = CLIPVectorizer(verbose=1)

        for file_path in tqdm(compatible_files):
            if remote:
                img_vector = json.loads(rpc.root.encode_image(file_path))[0]
            else:
                img_vector = img_vectorizer.encode_image(file_path)[0].tolist()

            self.db.add_file_embedding_vector(file_path, json.dumps(img_vector))
            self.db.conn.commit()
        print("Updating index...")
        if remote:
            rpc.root.update_image_index()
        else:
            img_vectorizer.update_index()

    def index_texts(self, rebuild=False, cache=False, cores: int = 0):
        """ Vectorize text files (needed for semantic search) """
        # TODO: index images
        # TODO: auto index on file addition (import)
        from .vectorizer import TextVectorizer, extract_clean_text, get_text_documents

        print("Vectorizing text documents...")
        remote = True
        if remote:
            try:
                rpc = rpyc.connect("localhost", 18861)
                rpc._config["sync_request_timeout"] = None  # Disable timeout
                print("Connected to DaemonService successfully")
            except ConnectionRefusedError:
                print("DaemonService connection failed, falling back to local execution...")
                remote = False
        if cache:
            print("Caching cleaned texts (database will grow big)")
        if rebuild:
            print("Rebuilding index")
            file_paths = self.db.get_vectorized_file_paths()
        else:
            file_paths = self.db.get_unvectorized_file_paths()
        i = 0
        compatible_files = get_text_documents(file_paths, verbose=True)
        min_words = 5
        min_word_length = 4
        args = []
        for file_path, file_type in compatible_files:
            args.append((file_path, file_type, cache, min_words, min_word_length))
        inference_tuples = []

        # Preprocess using multi-processing (default uses all available cores)
        if cores <= 0:
            n_cores = os.cpu_count()
        else:
            n_cores = cores
        pool = Pool(processes=n_cores)
        print(f"Preprocessing texts using {n_cores} cores...")
        with tqdm(total=len(compatible_files)) as t:
            for file_path, sentences in pool.imap_unordered(extract_clean_text, args):
                t.update(1)
                if sentences:
                    inference_tuples.append((file_path, sentences))
        print(f"Cleaned {len(inference_tuples)} text doc/s successfully")
        print("Starting inference...")
        # Compute embeddings
        if not remote:
            vectorizer = TextVectorizer(verbose=True)
        for file_path, sentences in tqdm(inference_tuples):
            if remote:
                document_vector = json.loads(rpc.root.compute_text_embedding(json.dumps(sentences)))
            else:
                document_vector = vectorizer.compute_text_embedding(sentences)
            if (
                document_vector is not None
                and type(document_vector) is list
                and len(document_vector) > 0
            ):
                self.db.add_file_embedding_vector(file_path, json.dumps(document_vector))
                self.db.conn.commit()
                i += 1
            else:
                print(type(document_vector))
                self.db.add_file_embedding_vector(file_path, json.dumps([]))
                self.db.conn.commit()
                print("Failed to parse file - skipping:", file_path)
        print(f"Vectorized {str(i)} file/s successfully")
        print("Updating index...")
        if remote:
            rpc.root.update_text_index()
        else:
            vectorizer.update_index()

    def search(self, *text_queries: str, path=False, top_k=10, score=False, _return=False):
        """ Execute a semantic search that returns best matching text documents """
        text_query = " . ".join(text_queries)
        try:
            rpc = rpyc.connect("localhost", 18861)
            results = rpc.root.search(text_query, path, top_k, score)
            for result in results:
                print(result)
            if len(result) == 0:
                print("No relevant files indexed...")
        except ConnectionRefusedError:
            from .vectorizer import TextVectorizer

            vectorizer = TextVectorizer()
            results = vectorizer.search(text_query, path, top_k, score)
        if _return:
            return results

    def add_auto_import_dir(self, path: str, index_images=False, index_texts=False):
        """ Add path for auto import directory (watched by daemon) """
        self.db.add_auto_import_directory(path, index_images, index_texts)

    def set_hypertagfs_dir(self, path: str):
        """ Set path for HyperTagFS directory """
        self.db.set_hypertagfs_dir(path)

    def mount(self, root_dir=None, parent_tag_id=None):
        """ Generate HyperTagFS: tag representation using symlinks """
        if root_dir is None:
            root_dir = self.root_dir
        root_path = Path(root_dir)

        if parent_tag_id is None:
            print("Updating HyperTagFS...")
            graph()
            os.makedirs(root_path / "Search Texts", exist_ok=True)
            os.makedirs(root_path / "Search Images", exist_ok=True)
            tag_ids_names = self.db.get_root_tag_ids_names()
            parent_file_paths_names = None
        else:
            tag_ids_names = self.db.get_tag_id_children_ids_names(parent_tag_id)
            parent_file_paths_names = set(
                self.db.get_file_paths_names_by_tag_id_shallow(parent_tag_id)
            )

        leaf_tag_ids = {tag_id[0] for tag_id in self.db.get_leaf_tag_ids()}
        dupes = dict()
        for tag_id, name in tag_ids_names:
            child_file_paths_names = set(self.db.get_file_paths_names_by_tag_id_shallow(tag_id))
            if parent_file_paths_names is None:
                file_paths_names = child_file_paths_names
            else:  # Intersect parent files with child
                file_paths_names = parent_file_paths_names.intersection(child_file_paths_names)
            if len(file_paths_names) > 0:
                underscore_root_tag_path = root_path / ("_" + name)
                root_tag_path = root_path / name
                if not root_tag_path.exists():
                    os.makedirs(underscore_root_tag_path, exist_ok=True)
                    move(underscore_root_tag_path, root_tag_path)  # Needed for daemon
                symlink_path = root_tag_path
                if tag_id not in leaf_tag_ids:
                    symlink_path = root_tag_path / "_files"
                    os.makedirs(symlink_path, exist_ok=True)
                for file_path, file_name in file_paths_names:
                    try:
                        filepath = Path(file_path)
                        current_symlink_path = symlink_path / file_name
                        if current_symlink_path.exists() and current_symlink_path.is_symlink():
                            existing_target = current_symlink_path.resolve()  # Get symlink target
                            if str(existing_target) != str(filepath):  # Duplicate?
                                dupe_i = dupes.get(current_symlink_path)
                                if dupe_i is None:
                                    dupes[current_symlink_path] = 1
                                    dupe_i = 2
                                dupes[current_symlink_path] += 1
                                current_symlink_path = symlink_path / (f"{dupe_i}-" + file_name)

                        os.symlink(filepath, current_symlink_path)
                    except FileExistsError:
                        pass
                self.mount(root_tag_path, tag_id)

    def auto_add_tags_from_path(
        self, file_path: Path, import_path_dirs: Set[str], verbose=False, keep_all=False
    ):
        file_path_tags = [p for p in str(file_path).split("/") if p not in import_path_dirs]
        if not keep_all:
            file_path_tags = file_path_tags[:-1]
        if verbose:
            print("Inferred tags:", file_path_tags)
        self.tag(
            file_path,
            "with",
            *file_path_tags,
            remount=False,
            add=False,
            commit=False,
        )
        for previous, current in zip(file_path_tags, file_path_tags[1:]):
            self.metatag(current, "with", previous, remount=False, commit=False)
        # print(file_path_tags, file_path.name)

    def import_tags(self, import_path: str, only_tags=False, verbose=False):
        """Import files with tags inferred from existing directory hierarchy
        (ignores hidden directories)"""
        file_paths = [p for p in list(Path(import_path).rglob("*")) if p.is_file()]
        # Remove files in hidden directories or in ignore list
        ignore_list = set(self.db.get_ignore_list())
        visible_file_paths = []
        for p in file_paths:
            is_hidden = False
            for pp in str(p).split("/"):
                if pp.startswith(".") or pp in ignore_list:
                    is_hidden = True
            if not is_hidden:
                visible_file_paths.append(p)
        print("Adding files...")
        if only_tags:
            added_file_paths = visible_file_paths
        else:
            added_file_paths = self.add(*visible_file_paths)
        import_path_dirs = set(str(import_path).split("/"))
        print("Adding tags...")
        for file_path in tqdm(added_file_paths):
            self.auto_add_tags_from_path(file_path, import_path_dirs, verbose)
        self.db.conn.commit()
        self.mount(self.root_dir)

    def remove(self, *file_names):
        """ Remove files """
        for file_name in tqdm(file_names):
            self.db.remove_file(file_name)
            remove_symlink(self.root_dir, file_name)
        self.mount(self.root_dir)

    def scrape(self, url, folder, timeout=1):
        config.setup_config(url, folder)
        wp = WebPage()
        wp.get(url)
        # start the saving process
        wp.save_complete()
        # join the sub threads
        for t in wp._threads:
            if t.is_alive():
                t.join(timeout)
        # location of the html file written
        return wp.file_path

    def add_url(self, url):
        webpages_regex = r"\S+.html|\S+.htm|\S+.php|\S+\/[^.]+"
        matches = re.findall(webpages_regex, url)
        if len(matches[0]) == len(url):
            web_pages_path = self.db.db_path / "web_pages"
            os.makedirs(web_pages_path, exist_ok=True)
            index_path = self.scrape(url, str(web_pages_path))
            return index_path
        else:
            downloads_path = self.db.db_path / "downloads"
            os.makedirs(downloads_path, exist_ok=True)
            file_name = url.split("/")[-1]
            file_path = downloads_path / file_name
            download_url(url, file_path)
            return file_path

    def add(self, *paths):
        """ Add file/s or URL/s"""
        added = []
        for path in tqdm(paths):
            try:
                if str(path).startswith("http"):
                    file_path = self.add_url(path)
                    self.db.add_file(os.path.abspath(file_path))
                    added.append(path)
                elif Path(path).is_file():
                    self.db.add_file(os.path.abspath(path))
                    added.append(path)
            except sqlite3.IntegrityError:
                pass
        self.db.conn.commit()
        print("Added", len(added), "new file/s")
        return added

    def tags(self, *file_names):
        """ Display all tags of file/s """
        tags = set()
        for file_name in file_names:
            tags.update(set(self.db.get_tags_by_file_name(file_name)))
        for tag in tags:
            print(tag)

    def show(self, mode="tags", path=False):
        """ Display all tags (default), indexed files (mode=index) or files """
        if mode == "files":
            names = self.db.get_files(path)
        elif mode == "index":
            names = self.db.get_vectorized_file_paths(path)
        elif mode == "tags":
            names = self.db.get_tags()
        for name in names:
            print(name)

    def query(self, *query, path=False, fuzzy=True, verbose=False):
        """Query files using set operands.
        Supported operands:
        and (intersection (default))
        or (union)
        minus (difference)
        """
        # TODO: Parse AST to support queries with brackets
        operands = {"and", "or", "minus"}
        tag_val = query[0].split("=")
        if len(tag_val) == 1:
            tag_query, value = (query[0], None)
        else:
            tag_query, value = (tag_val[0], tag_val[-1])
            value = value.replace("*", "%")
        results = set(self.db.get_files_by_tag(tag_query, path, fuzzy, value, verbose))
        current_operand = None
        for query_symbol in query[1:]:
            if query_symbol not in operands:
                tag_val = query_symbol.split("=")
                if len(tag_val) == 1:
                    tag_query, value = (query_symbol, None)
                else:
                    tag_query, value = (tag_val[0], tag_val[-1])
                    value = value.replace("*", "%")
                file_names = set(self.db.get_files_by_tag(tag_query, path, fuzzy, value, verbose))
                if current_operand == "or":
                    results = results.union(file_names)
                elif current_operand == "minus":
                    results = results - file_names
                else:
                    # and (intersection) is default operand
                    results = results.intersection(file_names)
            else:
                current_operand = query_symbol
        return results

    def tag(self, *args, remount=True, add=True, commit=True):
        """ Tag file/s with tag/s """
        # Parse arguments
        file_paths = []
        tags = []
        is_file_path = True
        for arg in args:
            if arg == "with":
                is_file_path = False
                continue
            if is_file_path:
                file_paths.append(arg)
            else:
                tag_val = arg.split("=")
                if len(tag_val) == 1:
                    tags.append((arg, None))
                else:
                    tags.append((tag_val[0], tag_val[-1]))
        if add:
            self.add(*file_paths)
        # Add tags to files
        for file_path in file_paths:
            for tag, value in tags:
                self.db.add_tag_to_file(tag, str(file_path), value)
                parent_tag_ids = self.db.get_parent_tag_ids_by_name(tag)
                # Add parent tags to file
                for parent_tag_id in parent_tag_ids:
                    self.db.add_tag_to_file(parent_tag_id, str(file_path))
        if commit:
            self.db.conn.commit()
        # Remount (everything is mounted)
        if remount:
            self.mount(self.root_dir)

    def untag(self, *args, remount=True, commit=True):
        """ Untag (remove tag/s of) file/s with tag/s """
        # Parse arguments
        file_names = []
        tags = []
        is_file_name = True
        for arg in args:
            if arg == "with":
                is_file_name = False
                continue
            if is_file_name:
                file_names.append(arg)
            else:
                tags.append(arg)
        # Remove tags
        for file_name in file_names:
            file_name = file_name.split("/")[-1]
            for tag in tags:
                self.db.remove_tag_from_file(tag, file_name)
            # print("Tagged", file_name, "with", tags)
        if commit:
            self.db.conn.commit()
        # TODO: Remove symlink (get all paths from metatags)
        # Remount (everything is mounted)
        if remount:
            self.mount(self.root_dir)

    def metatags(self, *tag_names):
        """ Display all metatags (parents) of tag/s """
        tags = set()
        for tag_name in tag_names:
            tags.update(set(self.db.get_meta_tags_by_tag_name(tag_name)))
        for tag in tags:
            print(tag)

    def metatag(self, *args, remount=True, commit=True):
        """ Tag tag/s with tag/s """
        # Parse arguments
        parent_tags = []
        tags = []
        is_parent_tag = False
        for arg in args:
            if arg == "with":
                is_parent_tag = True
                continue
            if is_parent_tag:
                parent_tags.append(arg)
            else:
                tags.append(arg)

        # Add meta tags
        for tag in tags:
            for parent_tag in parent_tags:
                self.db.add_parent_tag_to_tag(parent_tag, tag)
            # print("MetaTagged", tag, "with", parent_tags)

            # Add tags to files
            file_paths = self.db.get_files_by_tag(tag, show_path=True, fuzzy=True)
            for file_path in file_paths:
                self.db.add_tag_to_file(tag, str(file_path))
                parent_tag_ids = self.db.get_parent_tag_ids_by_name(tag)
                # Add parent tags to file
                for parent_tag_id in parent_tag_ids:
                    self.db.add_tag_to_file(parent_tag_id, str(file_path))

        if commit:
            self.db.conn.commit()

        for tag in tags:
            for parent_tag in parent_tags:
                # Remove parent_tag dir in all levels
                remove_dir(self.root_dir, parent_tag)
                # Remove tag dir in root level
                try:
                    rmtree(self.root_dir / tag)
                except Exception:  # nosec
                    pass  # Ignore if non existing
        # Remount (everything is mounted)
        if remount:
            self.mount(self.root_dir)

    def merge(self, tag_a, _into, tag_b):
        """ Merges all associations (files & tags) of tag_a into tag_b """
        print("Merging tag", tag_a, "into", tag_b)
        self.db.merge_tags(tag_a, tag_b)
        remove_dir(self.root_dir, tag_a)
        self.mount(self.root_dir)


def daemon(cpu=None):
    """ Start daemon process """
    print("Starting up daemon...")
    from .daemon import start
    from multiprocessing import Process, set_start_method

    set_start_method("spawn")
    p = Process(target=start, args=(cpu,))
    p.start()
    p.join()


def help():
    """ Get some help on how to use HyperTag """
    print(
        """
    README: https://github.com/SeanPedersen/HyperTag/blob/master/README.md
    Found a bug or simply feeling lost?
    Do not despair, you're not alone! Let us know at https://github.com/SeanPedersen/HyperTag/issues
    Chat and read about HyperTag using Matrix at https://bit.ly/2Xtz2Ff
    Print available CLI functions: hypertag
    Print available CLI flags for all commands: hypertag command --help
    """
    )


def main():
    ht = HyperTag()
    fire_cli = {
        "help": help,
        "add": ht.add,
        "remove": ht.remove,
        "import": ht.import_tags,
        "tag": ht.tag,
        "t": ht.tag,
        "untag": ht.untag,
        "metatag": ht.metatag,
        "tt": ht.metatag,
        "merge": ht.merge,
        "show": ht.show,
        "tags": ht.tags,
        "metatags": ht.metatags,
        "query": ht.query,
        "q": ht.query,
        "set_hypertagfs_dir": ht.set_hypertagfs_dir,
        "add_auto_import_dir": ht.add_auto_import_dir,
        "mount": ht.mount,
        "daemon": daemon,
        "graph": graph,
        "index": ht.index,
        "search_image": ht.search_image,
        "si": ht.search_image,
        "search": ht.search,
        "s": ht.search,
    }
    fire.Fire(fire_cli)


if __name__ == "__main__":
    main()



import os
import re
from pathlib import Path
import json
import threading
import torch
import rpyc  # type: ignore
from rpyc.utils.server import ThreadedServer  # type: ignore
from watchdog.observers import Observer  # type: ignore
from watchdog.events import FileSystemEventHandler  # type: ignore
from .persistor import Persistor
from .vectorizer import TextVectorizer, CLIPVectorizer
from .utils import update_symlink

text_vectorizer = None
image_vectorizer = None


class DaemonService(rpyc.Service):
    def on_connect(self, conn):
        print("New connection:", conn)

    def on_disconnect(self, conn):
        pass

    def exposed_compute_text_embedding(self, sentences_json):
        if text_vectorizer is not None:
            sentences = json.loads(sentences_json)
            return json.dumps(text_vectorizer.compute_text_embedding(sentences))

    def exposed_search(self, text_query: str, path=False, top_k=10, score=False):
        if text_vectorizer is not None:
            return text_vectorizer.search(text_query, path, top_k, score)

    def exposed_update_text_index(self):
        if text_vectorizer is not None:
            text_vectorizer.update_index()

    def exposed_encode_image(self, path: str):
        if image_vectorizer is not None:
            return json.dumps(image_vectorizer.encode_image(path).tolist())

    def exposed_search_image(self, text_query: str, path=False, top_k=10, score=False):
        if image_vectorizer is not None:
            return image_vectorizer.search_image(text_query, path, top_k, score)

    def exposed_update_image_index(self):
        if image_vectorizer is not None:
            image_vectorizer.update_index()


class AutoImportHandler(FileSystemEventHandler):
    def __init__(self, import_path):
        super().__init__()
        self.import_path = import_path
        with Persistor() as db:
            self.auto_index_images = db.get_auto_index_images(import_path)
            self.auto_index_texts = db.get_auto_index_texts(import_path)
        if self.auto_index_images:
            print("AutoImportHandler - Auto indexing images for", import_path)
        if self.auto_index_texts:
            print("AutoImportHandler - Auto indexing texts for", import_path)

    def on_moved(self, event):
        # TODO: Handle file rename / move
        super().on_moved(event)
        what = "directory" if event.is_directory else "file"
        print("AutoImportHandler - Moved", what, ": from", event.src_path, "to", event.dest_path)
        path = Path(event.dest_path)
        with Persistor() as db:
            file_id = db.get_file_id_by_path(event.src_path)

        from .hypertag import HyperTag

        ht = HyperTag()
        if file_id is None and path.is_file:  # Add new file
            print("AutoImportHandler - Adding file:", path)
            ht.add(str(path))
            import_path_dirs = set(str(self.import_path).split("/"))
            print("AutoImportHandler - Adding tags...")
            ht.auto_add_tags_from_path(path, import_path_dirs)
            ht.db.conn.commit()
            ht.mount(ht.root_dir)
            if self.auto_index_images:
                ht.index_images()
            if self.auto_index_texts:
                ht.index_texts()
        elif file_id is not None:  # Update existing path
            print("AutoImportHandler - Updating path & name for:", event.src_path, "to", path)
            old_name = event.src_path.split("/")[-1]
            with Persistor() as db:
                db.update_file_by_id(file_id, path)
                hypertagfs_path = db.get_hypertagfs_dir()
            # Update broken symlinks
            update_symlink(hypertagfs_path, old_name, path)
            # Add possible new tags
            import_path_dirs = set(str(event.src_path).split("/"))
            ht.auto_add_tags_from_path(path, import_path_dirs, verbose=True, keep_all=True)
            ht.db.conn.commit()
            # Remount
            ht.mount(ht.root_dir)

    def on_created(self, event):
        # Start auto import
        super().on_created(event)

        what = "directory" if event.is_directory else "file"
        print("AutoImportHandler - Created", what, event.src_path)
        path = Path(event.src_path)
        is_download_file = str(path).endswith(".crdownload")
        if path.is_file() and not is_download_file:  # Ignore download progress file
            from .hypertag import HyperTag

            ht = HyperTag()
            print("AutoImportHandler - Adding file:", path)
            ht.add(str(path))
            import_path_dirs = set(str(self.import_path).split("/"))
            print("AutoImportHandler - Adding tags...")
            ht.auto_add_tags_from_path(path, import_path_dirs)
            ht.db.conn.commit()
            ht.mount(ht.root_dir)
            if self.auto_index_images:
                ht.index_images()
            if self.auto_index_texts:
                ht.index_texts()

    def on_deleted(self, event):
        # TODO: Remove file and its tags
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"
        path = Path(event.src_path)
        print("AutoImportHandler - Deleted", what, path)


class HyperTagFSHandler(FileSystemEventHandler):
    def __init__(self):
        super().__init__()

    def on_moved(self, event):
        # Change tag association of moved file / dir
        super().on_moved(event)

        what = "directory" if event.is_directory else "file"
        print("Moved", what, ": from", event.src_path, "to", event.dest_path)

    def on_created(self, event):
        # DONE Interpret as query and populate with results as symlinks
        # TODO: Save created query dirs in dedicated table to recreate later
        super().on_created(event)

        what = "directory" if event.is_directory else "file"
        print("Created", what, event.src_path)
        path = Path(event.src_path)
        query = path.name
        if event.is_directory and not query.startswith("_"):
            from .hypertag import HyperTag

            ht = HyperTag()
            print("Populating with query results...")
            args = re.findall(r"'.*?'|\".*?\"|\S+", query)
            args = [e.replace('"', "").replace("'", "") for e in args]
            top_k = [int(a.split("=")[-1]) for a in args if a.startswith("top_k=")]
            if top_k:
                top_k = top_k[0]
            else:
                top_k = 10
            args = [a for a in args if not a.startswith("top_k=")]

            parent_tag_name = path.parent.name
            if parent_tag_name == "Search Texts":
                query_results = ht.search(*args, path=True, top_k=top_k, _return=True)
            elif parent_tag_name == "Search Images":
                query_results = ht.search_image(*args, path=True, top_k=top_k, _return=True)
            else:
                query_results = ht.query(*args, path=True)
            try:
                for fp in query_results:
                    file_path = Path(fp)
                    os.symlink(file_path, path / file_path.name)
            except Exception as ex:
                print("Failed:", ex)

    def on_deleted(self, event):
        """For symlink: remove tag
        For directory: remove tag associations"""
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"
        path = Path(event.src_path)
        print("Deleted", what, path)
        if event.is_directory and not path.name.startswith("_"):
            tag_name = path.name
            parent_tag_name = path.parent.name
            if parent_tag_name in {"HyperTagFS", "Search Texts", "Search Images"}:
                return
            print(
                "Removing parent tag association for:",
                tag_name,
                "Parent:",
                parent_tag_name,
            )
            with Persistor() as db:
                db.remove_parent_tag_from_tag(parent_tag_name, tag_name)
        else:  # Symlink
            tag_name = path.parent.name
            if tag_name == "_files":
                tag_name = path.parent.parent.name
            elif tag_name == "HyperTagFS":
                return
            file_name = path.name
            print("Removing tag:", tag_name, "from file:", file_name)
            with Persistor() as db:
                db.remove_tag_from_file(tag_name, file_name)

    def on_modified(self, event):
        super().on_modified(event)

        what = "directory" if event.is_directory else "file"
        print("Modified", what, ":", event.src_path)


def spawn_observer_thread(event_handler, path):
    print("Spawned observer for:", path)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    t = threading.Thread(target=observer.start)
    t.start()


def auto_importer():
    with Persistor() as db:
        auto_import_paths = db.get_auto_import_paths()
    print("Spawning Auto-Importer Threads...")
    for import_path in auto_import_paths:
        event_handler = AutoImportHandler(import_path)
        spawn_observer_thread(event_handler, import_path)


def watch_hypertagfs():
    with Persistor() as db:
        path = db.get_hypertagfs_dir()
    print("Spawning HyperTagFS Watching Thread...")
    event_handler = HyperTagFSHandler()
    spawn_observer_thread(event_handler, path)


def start(cpu):
    # Spawn Auto-Importer threads
    auto_importer()
    # Spawn HyperTagFS watch in thread
    watch_hypertagfs()

    cuda = torch.cuda.is_available()
    if cuda:
        print("CUDA runtime available")
    else:
        print("CUDA runtime not available (this might take a while)")
    print("Initializing TextVectorizer...")
    global text_vectorizer
    text_vectorizer = TextVectorizer(verbose=True)
    print("Initializing ImageVectorizer...")
    global image_vectorizer
    image_vectorizer = CLIPVectorizer(cpu, verbose=True)

    # IPC
    port = 18861
    # TODO: Investigate why Vectorize seems to be initialized twice...
    t = ThreadedServer(DaemonService, port=port)
    print(f"Starting DaemonService, listening on: localhost:{port}")
    t.start()


if __name__ == "__main__":
    print("Starting as standalone process")
    start()



from pathlib import Path
from igraph import Graph, plot  # type: ignore
from .persistor import Persistor


def graph(layout="fruchterman_reingold"):
    """ Visualize the HyperTag Graph (saved at HyperTagFS root dir) """
    graph_viz_name = "hypertag-graph.pdf"

    with Persistor() as db:
        hypertagfs_dir = Path(db.get_hypertagfs_dir())
        db.c.execute(
            """
            SELECT name, tag_id
            FROM tags
            """
        )
        tags = db.c.fetchall()
        tags.sort(key=lambda tup: tup[1])
        tag_ids = sorted([t[1] for t in tags])
        tags = [t[0] for t in tags]
        db.c.execute("SELECT * FROM tags_tags")
        raw_edges = db.c.fetchall()
        print(f"Visualizing HyperTag Graph (V:{len(tags)}, E:{len(raw_edges)})")

    edges: list = []
    edge_id_map: dict = {}
    i = 0
    for tid in tag_ids:
        edge_id_map[tid] = i
        i += 1

    edges = [(edge_id_map[a], edge_id_map[b]) for a, b in raw_edges]
    g = Graph(directed=True)
    g.add_vertices(tags)
    g.add_edges(edges)
    visual_style = {}
    visual_style["bbox"] = (800, 800)
    visual_style["vertex_size"] = 20  # type: ignore
    visual_style["margin"] = 50  # type: ignore
    visual_style["vertex_color"] = "yellow"  # type: ignore
    visual_style["layout"] = g.layout(layout)
    visual_style["vertex_label"] = g.vs["name"]
    plot(g, str(hypertagfs_dir / graph_viz_name), **visual_style)


if __name__ == "__main__":
    graph()



import os
import re
import json
import logging
from typing import Union, Tuple, List
from pathlib import Path
import filetype  # type: ignore
import torch
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor  # type: ignore
from PIL import Image  # type: ignore
from sentence_transformers import SentenceTransformer  # type: ignore
import hnswlib  # type: ignore
import PyPDF2  # type: ignore
import textract  # type: ignore
import wordninja  # type: ignore
from ftfy import fix_text  # type: ignore
from .persistor import Persistor
from .utils import download_url
from .tokenizer import SimpleTokenizer

logging.disable(logging.INFO)


class CLIPVectorizer:
    """ Multimodal vector space for images and texts powered by OpenAI's CLIP """

    def __init__(self, cpu=None, verbose=False):
        self.verbose = verbose
        self.device = "cuda" if torch.cuda.is_available() and not cpu else "cpu"
        if verbose:
            print("Using device:", self.device)
        TOKENIZER_URL = "https://openaipublic.azureedge.net/clip/bpe_simple_vocab_16e6.txt.gz"
        MODEL_URLS = {
            "cuda": "https://openaipublic.azureedge.net/clip/models/"
            + "40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
            "cpu": "https://battle.shawwn.com/sdb/models/ViT-B-32-cpu.pt",
        }
        db_path = Path.home() / ".config/hypertag/"
        clip_files_path = db_path / "CLIP-files"
        os.makedirs(clip_files_path, exist_ok=True)
        tokenizer_name = TOKENIZER_URL.split("/")[-1]
        if not Path(clip_files_path / tokenizer_name).is_file():
            print("Downloading tokenizer...")
            download_url(TOKENIZER_URL, clip_files_path / tokenizer_name)
        model_name = self.device + "-model.pt"
        if not Path(clip_files_path / model_name).is_file():
            print(f"Downloading CLIP {self.device.upper()} model...")
            download_url(MODEL_URLS[self.device], clip_files_path / model_name)

        self.model = torch.jit.load(str(clip_files_path / model_name), map_location=self.device)
        self.model = (
            self.model.eval()
            if torch.cuda.is_available() and not cpu
            else self.model.float().eval()
        )
        self.tokenizer = SimpleTokenizer(bpe_path=str(clip_files_path / tokenizer_name))
        input_resolution = self.model.input_resolution.item()
        self.preprocess = Compose(
            [
                Resize(input_resolution, interpolation=Image.BICUBIC),
                CenterCrop(input_resolution),
                ToTensor(),
            ]
        )

        self.image_mean = torch.tensor([0.48145466, 0.4578275, 0.40821073]).to(self.device)
        self.image_std = torch.tensor([0.26862954, 0.26130258, 0.27577711]).to(self.device)

        # Build or load index
        corpus_vectors, corpus_paths = self.get_image_corpus()
        index_dir = Path.home() / ".config/hypertag/index-files/"
        self.index_path = index_dir / "images.index"
        os.makedirs(index_dir, exist_ok=True)
        self.index = hnswlib.Index(space="cosine", dim=512)

        if self.index_path.exists():
            if self.verbose:
                print("Loading image index...")
            self.index.load_index(str(self.index_path), max_elements=len(corpus_vectors))
            self.update_index()
        else:
            # Create the HNSWLIB index
            if not corpus_vectors:
                return
            if self.verbose:
                print("Creating HNSWLIB image index...")
            self.index.init_index(max_elements=len(corpus_vectors), ef_construction=400, M=64)
            # Train the index to find a suitable clustering
            self.index.add_items(corpus_vectors, list(range(len(corpus_vectors))))
            if self.verbose:
                print("Saving index to:", self.index_path)
            self.index.save_index(str(self.index_path))
            # Update DB (set files as indexed)
            with Persistor() as db:
                db.set_indexed_by_file_paths(corpus_paths)
        # Controlling the recall by setting ef (lower is faster but more inaccuare)
        self.index.set_ef(50)  # ef should always be > top_k_hits

    def update_index(self):
        # Handle new vectorized elements
        with Persistor() as db:
            file_paths = db.get_unindexed_file_paths()
            compatible_files = get_image_files(file_paths)
            corpus = db.get_file_embedding_vectors(compatible_files)
        new_corpus_paths = []
        new_corpus_vectors = []
        for doc_path, embedding_vector in corpus:
            new_corpus_vectors.append(json.loads(embedding_vector))
            new_corpus_paths.append(doc_path)
        len_new = len(new_corpus_vectors)
        len_old = self.index.get_current_count()
        new_total_size = len_old + len_new
        if self.verbose:
            print("CURRENT INDEXED FILES:", len_old)
            print("NEW UNINDEXED FILES:", len_new)
            print("NEW TOTAL SIZE:", new_total_size)
        if new_corpus_vectors:
            # Add unindexed vectors to index
            self.index.resize_index(new_total_size)
            new_ids = list(range(len_old, new_total_size))
            self.index.add_items(
                new_corpus_vectors,
                new_ids,
            )
            if self.verbose:
                print("Saving updated index to:", self.index_path)
            self.index.save_index(str(self.index_path))
            # Update DB (set files as indexed)
            with Persistor() as db:
                db.set_indexed_by_file_paths(new_corpus_paths)

    def get_image_corpus(self):
        # Retrieve vectorized image vectors and paths
        with Persistor() as db:
            file_paths = db.get_vectorized_file_paths()
            compatible_files = get_image_files(file_paths)
            corpus = db.get_file_embedding_vectors(compatible_files)

        corpus_paths = []
        corpus_vectors = []
        for doc_path, embedding_vector in corpus:
            corpus_vectors.append(json.loads(embedding_vector))
            corpus_paths.append(doc_path)
        return corpus_vectors, corpus_paths

    def search_image(self, text_query: str, path, top_k, score):
        query_vector = None
        # Check if text_query is an image file path
        file_path = Path(text_query)
        if file_path.exists() and file_path.is_file():
            file_type_guess = filetype.guess(str(file_path))
            if file_type_guess and file_type_guess.extension in {"jpg", "png"}:
                query_vector = self.encode_image(str(file_path))

        corpus_vectors, corpus_paths = self.get_image_corpus()
        image_features = torch.Tensor(corpus_vectors)
        image_features /= image_features.norm(dim=-1, keepdim=True)

        # Encode text query
        if query_vector is None:
            query_vector = self.encode_text(text_query)
        query_vector /= query_vector.norm(dim=-1, keepdim=True)
        # Indexed top-k nearest neighbor query
        corpus_ids, distances = self.index.knn_query(query_vector.cpu(), k=top_k)
        # Print results
        results = []
        for corpus_id, score_value in zip(corpus_ids[0], distances[0]):
            match_file_path = str(corpus_paths[int(corpus_id)])
            file_name = match_file_path.split("/")[-1]
            if path:
                file_name = match_file_path
            if score:
                result = f"{file_name} ({score_value:.4f})"
                results.append(result)
                print(result)
            else:
                results.append(file_name)
                print(file_name)
        return results

    def encode_image(self, path: str):
        image = Image.open(path).convert("RGB")
        image = self.preprocess(image)
        image.unsqueeze_(0)
        image_input = image.to(self.device)
        image_input -= self.image_mean[:, None, None]
        image_input /= self.image_std[:, None, None]
        with torch.no_grad():
            image_features = self.model.encode_image(image_input).float()
        return image_features

    def encode_text(self, text: str):
        sot_token = self.tokenizer.encoder["<|startoftext|>"]
        eot_token = self.tokenizer.encoder["<|endoftext|>"]
        text_token = self.tokenizer.encode(text)
        text_tokens = [[sot_token] + text_token + [eot_token]]
        text_input = torch.zeros(len(text_tokens), self.model.context_length, dtype=torch.long)

        for i, tokens in enumerate(text_tokens):
            text_input[i, : len(tokens)] = torch.tensor(tokens)

        text_input = text_input.to(self.device)

        with torch.no_grad():
            text_features = self.model.encode_text(text_input).float()
        return text_features


class TextVectorizer:
    def __init__(self, verbose=False):
        self.verbose = verbose
        os.environ["TOKENIZERS_PARALLELISM"] = "true"
        model_name = "stsb-distilbert-base"
        self.model = SentenceTransformer(model_name)

        # Build or load index
        corpus_vectors, corpus_paths = self.get_text_corpus()
        index_dir = Path.home() / ".config/hypertag/index-files/"
        self.index_path = index_dir / "texts.index"
        os.makedirs(index_dir, exist_ok=True)
        self.index = hnswlib.Index(space="cosine", dim=768)

        if self.index_path.exists():
            if self.verbose:
                print("Loading text index...")
            self.index.load_index(str(self.index_path), max_elements=len(corpus_vectors))
            self.update_index()
        else:
            # Create the HNSWLIB index
            if not corpus_vectors:
                return
            if self.verbose:
                print("Creating HNSWLIB text index...")
            self.index.init_index(max_elements=len(corpus_vectors), ef_construction=400, M=64)
            # Then we train the index to find a suitable clustering
            self.index.add_items(corpus_vectors, list(range(len(corpus_vectors))))
            if self.verbose:
                print("Saving index to:", self.index_path)
            self.index.save_index(str(self.index_path))
            # Update DB (set files as indexed)
            with Persistor() as db:
                db.set_indexed_by_file_paths(corpus_paths)
        # Controlling the recall by setting ef (lower is faster but more inaccuare)
        self.index.set_ef(50)  # ef should always be > top_k_hits

    def update_index(self):
        # Handle new vectorized elements
        with Persistor() as db:
            file_paths = db.get_unindexed_file_paths()
            compatible_files = [p for p, _t in get_text_documents(file_paths)]
            corpus = db.get_file_embedding_vectors(compatible_files)
        new_corpus_paths = []
        new_corpus_vectors = []
        for doc_path, embedding_vector in corpus:
            new_corpus_vectors.append(json.loads(embedding_vector))
            new_corpus_paths.append(doc_path)
        len_new = len(new_corpus_vectors)
        len_old = self.index.get_current_count()
        new_total_size = len_old + len_new
        if self.verbose:
            print("CURRENT INDEXED FILES:", len_old)
            print("NEW UNINDEXED FILES:", len_new)
            print("NEW TOTAL SIZE:", new_total_size)
        if new_corpus_vectors:
            # Add unindexed vectors to index
            self.index.resize_index(new_total_size)
            new_ids = list(range(len_old, new_total_size))
            self.index.add_items(
                new_corpus_vectors,
                new_ids,
            )
            if self.verbose:
                print("Saving updated index to:", self.index_path)
            self.index.save_index(str(self.index_path))
            # Update DB (set files as indexed)
            with Persistor() as db:
                db.set_indexed_by_file_paths(new_corpus_paths)

    def get_text_corpus(self):
        # Returns text paths and embedding vectors
        with Persistor() as db:
            text_files = db.get_files(show_path=True)
        text_document_tuples = get_text_documents(text_files)
        text_document_paths = [path for path, _file_type in text_document_tuples]
        with Persistor() as db:
            corpus = db.get_file_embedding_vectors(text_document_paths)
        corpus_paths = []
        corpus_vectors = []
        for doc_path, embedding_vector in corpus:
            corpus_vectors.append(json.loads(embedding_vector))
            corpus_paths.append(doc_path)
        return corpus_vectors, corpus_paths

    def compute_text_embedding(self, sentences: List[List[str]]) -> List[float]:
        for i, s in enumerate(sentences):
            if len(s) == 1:
                sentences[i] = 2 * sentences[i]  # Needed for transformer model
        sentence_vectors = self.model.encode(sentences, show_progress_bar=False)
        if sentence_vectors.shape[0] > 0:
            return torch.Tensor(sentence_vectors.mean(axis=0)).tolist()
        else:
            return torch.Tensor([sentence_vectors]).tolist()

    def search(self, text_query: str, path=False, top_k=10, score=False):
        """ Execute a semantic search that returns best matching text documents """
        # Parse query: duplicate words marked with * (increases search weight)
        parsed_query = []
        rex = re.compile(r"\*+\w+")
        args = re.findall(r"\*+'.*?'|\*+\".*?\"|\S+", text_query)

        for w in args:
            if w.startswith("*"):
                w = w.replace('"', "").replace("'", "")
                matches = rex.search(w)
                if matches:
                    replicate_n = matches.group().count("*")
                    for _ in range(replicate_n + 1):
                        parsed_query.append(w[replicate_n:])
            else:
                parsed_query.append(w)
        parsed_text_query = " ".join(parsed_query)
        corpus_vectors, corpus_paths = self.get_text_corpus()

        sentence_query = clean_transform(parsed_text_query, 0, 1)
        query_vector = self.compute_text_embedding(sentence_query)
        # Indexed top-k nearest neighbor query
        corpus_ids, distances = self.index.knn_query(query_vector, k=top_k)

        results = []
        for corpus_id, score_value in zip(corpus_ids[0], distances[0]):
            file_path = corpus_paths[corpus_id]
            file_name = file_path.split("/")[-1]
            if path:
                file_name = file_path
            if score:
                result = f"{file_name} ({score_value:.4f})"
                results.append(result)
                print(result)
            else:
                results.append(file_name)
                print(file_name)
        return results


def get_image_files(file_paths, verbose=False):
    # Keep only images (JPG)
    doc_types = {"jpg", "png"}
    if verbose:
        print("Supported image file types:", doc_types)
    compatible_files = []
    for file_path in file_paths:
        file_type_guess = filetype.guess(str(file_path))
        if file_type_guess is None:
            continue
        file_type = file_type_guess.extension.lower()
        if file_type in doc_types:
            compatible_files.append(str(file_path))
    return compatible_files


def get_text_documents(file_paths: List[str], verbose=False) -> List[Tuple[str, str]]:
    # TODO: Add markdown support
    doc_types = {
        "pdf",
        "epub",
        "doc",
        "docx",
        "html",
        "json",
        "rtf",
        "txt",
        "xls",
        "xlsx",
        "ps",
        "pptx",
        "odt",
        "eml",
        "msg",
    }
    if verbose:
        print("Supported text file types:", doc_types)
    # Keep only text files
    compatible_files = []
    for file_path in file_paths:
        file_type_guess = filetype.guess(str(file_path))
        if file_type_guess is None:
            file_name = file_path.split("/")[-1]
            name_parts = file_name.split(".")
            if len(name_parts) > 1 and len(name_parts[-1]) < 6:
                file_type = name_parts[-1]
            else:
                continue
        else:
            file_type = file_type_guess.extension.lower()
        if file_type in doc_types:
            compatible_files.append((str(file_path), file_type))
    return compatible_files


def extract_clean_text(args: Tuple[str, str, bool, int, int]) -> Tuple[str, List[List[str]]]:
    file_path, file_type, cache, min_words, min_word_length = args
    if cache:
        with Persistor() as db:
            # Use saved cleaned text if available (cache)
            # TODO: Evaluate thoroughly if storing all tokens is worth it
            # Hunch: Yes, we can store text tokens and a single mapping (token : word)
            sentences_string = db.get_clean_text_of_file(file_path)
            if sentences_string is not None:
                sentences = [s.split(" ") for s in json.loads(sentences_string)]
                return str(file_path), sentences

    text = extract_text(file_path, file_type)
    sentences = []
    if text:
        sentences = clean_transform(text, min_words, min_word_length)
        if cache:
            with Persistor() as db:
                # Save cleaned text
                db.add_clean_text_to_file(file_path, json.dumps([" ".join(s) for s in sentences]))
    else:
        with Persistor() as db:
            # Mark as not parseable
            print("NOT PARSEABLE")
            db.add_clean_text_to_file(file_path, json.dumps([]))
    return str(file_path), sentences


def extract_text(file_path_: str, file_type: str) -> str:
    file_path = Path(file_path_)
    # print(f"Parsing: {file_path} as type: {file_type}")
    if file_type == "pdf":
        text = get_pdf_text(file_path)
    else:
        try:
            text = str(textract.process(file_path))
        except Exception:
            # print("Exception while parsing:", ex)
            text = ""  # Failed to parse...
    return text


def clean_transform(text: str, min_words: int, min_word_length: int) -> List[List[str]]:
    text = fix_text(text, normalization="NFKC")
    text = text.replace("\n", " ").replace("\t", " ")
    text = re.sub("[^A-Za-z0-9 .']+", "", text)  # Remove all special chars but space . '
    sentences = [s for s in text.split(".") if s]

    ninja_sentences = []
    for s in sentences:
        ninja_sentence = []
        for w in s.split(" "):
            if len(w) > 23:  # Words longer than 23 chars are likely missing spaces
                for nw in wordninja.split(w):  # Split non space seperated word groups up
                    if len(nw) > min_word_length:
                        ninja_sentence.append(nw)
            else:
                if len(w) > min_word_length:
                    ninja_sentence.append(w)
        ninja_sentences.append(ninja_sentence)
    # Remove duplicates
    ninja_sentence_set = {" ".join(s) for s in ninja_sentences}

    long_sentences = [s.split(" ") for s in ninja_sentence_set if len(s.split(" ")) > min_words]
    return long_sentences


def get_pdf_text(file_path: Union[Path, str]) -> str:
    text = ""
    try:
        text = textract.process(file_path, encoding="utf8").decode("utf-8")
    except Exception:
        try:
            text = get_pypdf_text(file_path)
        except Exception as ex:
            print("Exception while parsing:", ex)
    if not text:
        print("Falling back to Tesseract OCR parsing (get a tea...) for:", file_path)
        try:
            text = textract.process(file_path, method="tesseract").decode("utf-8")
        except Exception as ex:
            print("OCR failed:", ex)
            text = ""
    return text


def get_pypdf_text(file_path: Union[Path, str]) -> str:
    text = ""
    parsed = 0
    failed = 0
    with open(str(file_path), mode="rb") as f:
        reader = PyPDF2.PdfFileReader(f, strict=False)
        for page in reader.pages:
            try:
                if failed > 42:
                    print("Stopping to parse after 42 failed pages...")
                    return ""
                parsed += 1
                text += " " + page.extractText()
            except Exception:
                # print("failed to parse page", parsed)
                failed += 1
    return text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hypertag']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.1.0,<9.0.0',
 'PyPDF2>=1.26.0,<2.0.0',
 'cairocffi>=1.2.0,<2.0.0',
 'filetype>=1.0.7,<2.0.0',
 'fire>=0.3.1,<0.4.0',
 'ftfy>=5.8,<6.0',
 'fuzzywuzzy[speedup]>=0.18.0,<0.19.0',
 'hnswlib>=0.4.0,<0.5.0',
 'python-igraph>=0.8.3,<0.9.0',
 'pywebcopy>=6.3.0,<7.0.0',
 'rpyc>=5.0.0,<6.0.0',
 'sentence-transformers>=0.4.1,<0.5.0',
 'textract>=1.6.3,<2.0.0',
 'torchvision>=0.8.2,<0.9.0',
 'tqdm>=4.55.0,<5.0.0',
 'watchdog>=1.0.2,<2.0.0',
 'wordninja>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'hypertag',
    'version': '0.6.2.5',
    'description': 'File organization made for humans',
    'long_description': '# HyperTag\n\nHyperTag helps humans intuitively express how they think about their files using tags and machine learning. Represent how you think using tags. Find what you look for using semantic search for your text documents (yes, even PDF\'s) and images. Instead of introducing proprietary file formats like other existing file organization tools, HyperTag just smoothly layers on top of your existing files without any fuss.\n\n**Objective Function**: Minimize time between a thought and access to all relevant files.\n\nAccompanying blog post: https://blog.neotree.uber.space/posts/hypertag-file-organization-made-for-humans\n\n<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON\'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n**Table of Contents**\n- [HyperTag](#hypertag)\n  - [Install](#install)\n  - [Community](#community)\n  - [Overview](#overview)\n  - [CLI Functions](#cli-functions)\n    - [Import existing directory recursively](#import-existing-directory-recursively)\n    - [Add file/s or URL/s manually](#add-files-or-urls-manually)\n    - [Tag file/s  (with values)](#tag-files--with-values)\n    - [Untag file/s](#untag-files)\n    - [Tag a tag](#tag-a-tag)\n    - [Merge tags](#merge-tags)\n    - [Query using Set Theory](#query-using-set-theory)\n    - [Index supported image and text files](#index-supported-image-and-text-files)\n    - [Semantic search for text files](#semantic-search-for-text-files)\n    - [Semantic search for image files](#semantic-search-for-image-files)\n    - [Start HyperTag Daemon](#start-hypertag-daemon)\n    - [Print all tags of file/s](#print-all-tags-of-files)\n    - [Print all metatags of tag/s](#print-all-metatags-of-tags)\n    - [Print all tags](#print-all-tags)\n    - [Print all files](#print-all-files)\n    - [Visualize HyperTag Graph](#visualize-hypertag-graph)\n    - [Generate HyperTagFS](#generate-hypertagfs)\n    - [Add directory to auto import list](#add-directory-to-auto-import-list)\n    - [Set HyperTagFS directory path](#set-hypertagfs-directory-path)\n  - [Architecture](#architecture)\n  - [Development](#development)\n  - [Inspiration](#inspiration)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n## Install\nAvailable on [PyPI](https://pypi.org/project/hypertag/)\n\n`$ pip install hypertag` (supports both CPU only & CUDA accelerated execution!)\n\n## Community\nJoin the HyperTag [matrix chat room](https://matrix.to/#/#hypertag:matrix.neotree.uber.space?via=matrix.neotree.uber.space) to stay up to date on the latest developments or to ask for help.\n\n## Overview\nHyperTag offers a slick CLI but more importantly it creates a directory called ```HyperTagFS``` which is a file system based representation of your files and tags using symbolic links and directories.\n\n**Directory Import**: Import your existing directory hierarchies using ```$ hypertag import path/to/directory```. HyperTag converts it automatically into a tag hierarchy using metatagging.\n\n**Semantic Text & Image Search  (Experimental)**: Search for **images** (jpg, png) and **text documents** (yes, even PDF\'s) content with a simple text query. Text search is powered by the awesome [Sentence Transformers](https://github.com/UKPLab/sentence-transformers) library. Text to image search is powered by OpenAI\'s [CLIP model](https://openai.com/blog/clip/). Currently only English queries are supported.\n\n**HyperTag Daemon  (Experimental)**: Monitors `HyperTagFS` and directories added to the auto import list for user changes (see section "Start HyperTag Daemon" below). Also spawns the DaemonService which speeds up semantic search significantly.\n\n**Fuzzy Matching Queries**: HyperTag uses fuzzy matching to minimize friction in the unlikely case of a typo.\n\n**File Type Groups**: HyperTag automatically creates folders containing common files (e.g. Images: jpg, png, etc., Documents: txt, pdf, etc., Source Code: py, js, etc.), which can be found in ```HyperTagFS```.\n\n**HyperTag Graph**: Quickly get an overview of your HyperTag Graph! HyperTag visualizes the metatag graph on every change and saves it at `HyperTagFS/hypertag-graph.pdf`.\n\n![HyperTag Graph Example](https://raw.githubusercontent.com/SeanPedersen/HyperTag/master/images/hypertag-graph.jpg)\n\n## CLI Functions\n\n### Import existing directory recursively\nImport files with tags inferred from the existing directory hierarchy.\n\n```$ hypertag import path/to/directory```\n\n### Add file/s or URL/s manually\n\n```$ hypertag add path/to/file https://github.com/SeanPedersen/HyperTag```\n\n### Tag file/s  (with values)\nManually tag files. Shortcut: `$ hypertag t`\n\n```$ hypertag tag humans/*.txt with human "Homo Sapiens"```\n\nAdd a value to a file\'s tag:\n\n```$ hypertag tag sean.txt with name="Sean Pedersen"```\n\n### Untag file/s\nManually remove tag/s from file/s.\n\n```$ hypertag untag humans/*.txt with human "Homo Sapiens"```\n\n### Tag a tag\nMetatag tag/s to create tag hierarchies. Shortcut: `$ hypertag tt`\n\n```$ hypertag metatag human with animal```\n\n### Merge tags\nMerge all associations (files & tags) of tag A into tag B.\n\n```$ hypertag merge human into "Homo Sapiens"```\n\n### Query using Set Theory\nPrint file names of the resulting set matching the query. Queries are composed of tags (with values) and operands. Tags are fuzzy matched for convenience. Nesting is currently not supported, queries are evaluated from left to right.\n<br>Shortcut: `$ hypertag q`\n\nQuery with a value using a wildcard: ```$ hypertag query name="Sean*"```<br>\nPrint paths: ```$ hypertag query human --path```<br>\nPrint fuzzy matched tag: ```$ hypertag query man --verbose``` <br>\nDisable fuzzy matching: ```$ hypertag query human --fuzzy=0```\n\nDefault operand is AND (intersection): <br>\n```$ hypertag query human name="Sean*"``` is equivalent to ```$ hypertag query human and name="Sean*"```\n\nOR (union): <br>\n```$ hypertag query human or "Homo Sapiens"```\n\nMINUS (difference): <br>\n```$ hypertag query human minus "Homo Sapiens"```\n\n### Index supported image and text files\nOnly indexed files can be searched.\n\n```$ hypertag index```\n\nTo parse even unparseable PDF\'s, install tesseract: `# pacman -S tesseract tesseract-data-eng`\n\nIndex only image files: ```$ hypertag index --image```<br>\nIndex only text files: ```$ hypertag index --text```\n\n### Semantic search for text files\nPrint text file names sorted by matching score.\nPerformance benefits greatly from running the HyperTag daemon.\n<br>Shortcut: `$ hypertag s`\n\n```$ hypertag search "your important text query" --path --score --top_k=10```\n\n### Semantic search for image files\nPrint image file names sorted by matching score.\nPerformance benefits greatly from running the HyperTag daemon.\n<br>Shortcut: `$ hypertag si`\n\nText to image:\n```$ hypertag search_image "your image content description" --path --score --top_k=10```\n\nImage to image:\n```$ hypertag search_image "path/to/image.jpg" --path --score --top_k=10```\n\n### Start HyperTag Daemon\nStart daemon process with triple functionality:\n- Watches `HyperTagFS` directory for user changes\n  - Maps file (symlink) and directory deletions into tag / metatag removal/s\n  - On directory creation: Interprets name as set theory tag query and automatically populates it with results\n  - On directory creation in `Search Images` or `Search Texts`: Interprets name as semantic search query (add top_k=42 to limit result size) and automatically populates it with results\n- Watches directories on the auto import list for user changes:\n  - Maps file changes (moves & renames) to DB\n  - On file creation: Adds new file/s with inferred tag/s\n- Spawns DaemonService to load and expose models used for semantic search, speeding it up significantly\n\n```$ hypertag daemon```\n\n### Print all tags of file/s\n\n```$ hypertag tags filename1 filename2```\n\n### Print all metatags of tag/s\n\n```$ hypertag metatags tag1 tag2```\n\n### Print all tags\n\n```$ hypertag show```\n\n### Print all files\n\nPrint names:\n```$ hypertag show files```\n\nPrint paths:\n```$ hypertag show files --path```\n\n### Visualize HyperTag Graph\nVisualize the metatag graph hierarchy (saved at HyperTagFS root).\n\n```$ hypertag graph```\n\nSpecify [layout algorithm](https://igraph.org/python/doc/tutorial/tutorial.html#layout-algorithms) (default: fruchterman_reingold):\n\n```$ hypertag graph --layout=kamada_kawai```\n\n### Generate HyperTagFS\nGenerate file system based representation of your files and tags using symbolic links and directories.\n\n```$ hypertag mount```\n\n### Add directory to auto import list\nDirectories added to the auto import list will be monitored by the daemon for new files or changes.\n\n```$ hypertag add_auto_import_dir path/to/directory```\n\n### Set HyperTagFS directory path\nDefault is the user\'s home directory.\n\n```$ hypertag set_hypertagfs_dir path/to/directory```\n\n## Architecture\n- Python and it\'s vibrant open-source community power HyperTag\n- Many other awesome open-source projects make HyperTag possible (listed in `pyproject.toml`)\n- SQLite3 serves as the meta data storage engine (located at `~/.config/hypertag/hypertag.db`)\n- Added URLs are saved in `~/.config/hypertag/web_pages` for websites, others in `~/.config/hypertag/downloads`\n- Symbolic links are used to create the HyperTagFS directory structure\n- Semantic Search: boosted using [hnswlib](https://github.com/nmslib/hnswlib/)\n  - Text to text search is powered by the awesome [DistilBERT](https://arxiv.org/abs/1910.01108)\n  - Text to image & image to image search is powered by OpenAI\'s impressive [CLIP model](https://openai.com/blog/clip/)\n\n## Development\n- Find prioritized issues here: [TODO List](https://github.com/SeanPedersen/HyperTag/projects/1)\n- Pick an issue and comment how you plan to tackle it before starting out, to make sure no dev time is wasted.\n- Clone repo: ```$ git clone https://github.com/SeanPedersen/HyperTag.git```\n- `$ cd HyperTag/`\n- Install [Poetry](https://python-poetry.org/docs/#installation)\n- Install dependencies: `$ poetry install`\n- Activate virtual environment: `$ poetry shell`\n- Run all tests: ```$ pytest -v```\n- Run formatter: ```$ black hypertag/```\n- Run linter: ```$ flake8```\n- Run type checking: ```$ mypy **/*.py```\n- Run security checking: ```$ bandit --exclude tests/ -r .```\n- Run HyperTag: ```$ python -m hypertag```\n\n## Inspiration\n\n**What is the point of HyperTag\'s existence?**<br/>\nHyperTag offers many unique features such as the import, semantic search for images and texts, graphing and fuzzy matching functions that make it very convenient to use. All while HyperTag\'s code base staying tiny at <2000 LOC in comparison to TMSU (>10,000 LOC) and SuperTag (>25,000 LOC), making it easy to hack on.\n\nThis project is partially inspired by these open-source projects:\n- [TMSU](https://github.com/oniony/TMSU): Written in Go\n- [SuperTag](https://github.com/amoffat/supertag): Written in Rust\n',
    'author': 'Sean',
    'author_email': 'sean-p-96@hotmail.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SeanPedersen/HyperTag',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
